"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});
/* Helpers to control access permissions of favorites and interpretations.
 *
 * A favorite sharing must be a superset of all its interpretations sharings.
 * An interpretation sharing must be a subset of its parent object sharing.

 Permissions checked: publicAccess, userGroupAccesses and userAccesses.
 */

var accesses = {
    "--------": "private",
    "r-------": "read",
    "rw------": "read-write"
};

var permissionFields = ["publicAccess", "userAccesses", "userGroupAccesses"];

var toArray = function toArray(obj) {
    return obj || [];
};

var checkValidationCondition = exports.checkValidationCondition = function checkValidationCondition(validator, oldValue, newValue, showError) {
    var oldValueValidates = oldValue ? validator(oldValue) : true;
    var newValueValidates = validator(newValue);

    if (oldValueValidates && newValueValidates) {
        return true;
    } else if (oldValueValidates && !newValueValidates) {
        showError && showError();
        return false;
    } else if (!oldValueValidates && newValueValidates) {
        return true;
    } else if (!oldValueValidates && !newValueValidates) {
        showError && showError();
        return true;
    }
};

var isObjectAccessible = function isObjectAccessible(favoriteAccess, objectAccess) {
    return !(accesses[favoriteAccess] === "private" && accesses[objectAccess] !== "private");
};

var validateFieldAccess = exports.validateFieldAccess = function validateFieldAccess(modelName, favorite, showError, field, oldValue, newValue) {
    var modelKey = modelName === "interpretation" ? modelName : "favorite";
    var interpretations = toArray(favorite.interpretations);
    var predicate = void 0;

    switch (modelKey + "-" + field) {
        /* Interpretations validations: check that its sharing is a subset of the favorite sharing */
        case "interpretation-publicAccess":
            predicate = function predicate(value) {
                return isObjectAccessible(favorite[field], value);
            };
            break;
        case "interpretation-userAccesses":
        case "interpretation-userGroupAccesses":
            var favoritePermissionIds = new Set(toArray(favorite[field]).map(function (p) {
                return p.id;
            }));
            predicate = function predicate(permissions) {
                return toArray(permissions).every(function (permission) {
                    return favoritePermissionIds.has(permission.id);
                });
            };
            break;

        /* Favorite validations: check that all its interpretation sharings are a subset of this favorite sharing */
        case "favorite-publicAccess":
            predicate = function predicate(value) {
                return interpretations.every(function (interpretation) {
                    return isObjectAccessible(value, interpretation[field]);
                });
            };
            break;
        case "favorite-userAccesses":
        case "favorite-userGroupAccesses":
            predicate = function predicate(favoritePermissions) {
                var favoritePermissionIds = new Set(toArray(favoritePermissions).map(function (p) {
                    return p.id;
                }));
                return interpretations.every(function (interpretation) {
                    return toArray(interpretation[field]).every(function (interpretationPermission) {
                        return favoritePermissionIds.has(interpretationPermission.id);
                    });
                });
            };
            break;
        default:
            throw new Error("[validateFieldAccess] unknown field key: " + modelKey);
    }

    return checkValidationCondition(predicate, oldValue, newValue, showError);
};

var validateSharing = exports.validateSharing = function validateSharing(modelName, favorite, sharing, showError) {
    if (!favorite || !sharing) {
        return true;
    } else {
        var isValid = permissionFields.every(function (field) {
            return validateFieldAccess(modelName, favorite, null, field, null, sharing[field]);
        });
        if (!isValid) showError();
        return isValid;
    }
};
//# sourceMappingURL=permissions.js.map